﻿using System;
using System.IO;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Microsoft.Win32; // Registry
using System.Runtime.InteropServices;
using System.Reflection;
using System.Text;

namespace FileDocumentControlLibrary {

  [DefaultEvent("NewDocument")]
  public partial class FileDocument : Component, ISupportInitialize {

    private object data = null;
    private bool dirty = false;
    private string fileName = "";
    private Form host = null;
    private string defaultFileName = "Untitled";
    private string fileExtension = null;
    private string fileExtensionDesc = null;
    private bool registerFileExtension = false;
    private string filter = null;
    private string iconResourceIndex = "";
    private IFormatter formatter = new BinaryFormatter();
    private ToolStripMenuItem fileNewMenuItem = null;
    private ToolStripMenuItem fileOpenMenuItem = null;
    private ToolStripMenuItem fileSaveMenuItem = null;
    private ToolStripMenuItem fileSaveAsMenuItem = null;
    private ToolStripMenuItem fileSaveCopyAsMenuItem = null;
    private ToolStripMenuItem fileCloseMenuItem = null;
    private ToolStripButton fileNewToolStripButton = null;
    private ToolStripButton fileOpenToolStripButton = null;
    private ToolStripButton fileSaveToolStripButton = null;
    private ToolStripButton fileSaveAsToolStripButton = null;
    private ToolStripButton fileSaveCopyAsToolStripButton = null;
    private ToolStripButton fileCloseToolStripButton = null;

    public FileDocument() {
      InitializeComponent();
    }
    public FileDocument(IContainer container) {
      // Integrate with container
      if( container != null ) container.Add(this);

      // Required for Designer support
      InitializeComponent();
    }

    #region ISupportInitialize Members
    public void BeginInit() {}

    public void EndInit() {
      // Called after container is done initializing all properties
      CheckRegisterFileExtension();
    }

    #endregion

    #region Public Events

    [Category("Action")]
    [Description("Occurs when either FileDocument is instantiated, or a new document is created with the New method.")]
    public event EventHandler NewDocument;

    [Category("Action")]
    [Description("Occurs when a file is passed through the command line or if FileOpenMenuItem is clicked, if handled.")]
    public event SerializeDocumentEventHandler ReadDocument;

    [Category("Action")]
    [Description("Occurs when FileSaveMenuItem is clicked, if handled.")]
    public event SerializeDocumentEventHandler WriteDocument;

    [Category("Property Changed")]
    [Description("Occurs when the value of the Document property is changed.")]
    public event EventHandler DocumentChanged;

    #endregion

    #region Public Properties
    [DefaultValue(null)]
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public object Data {
      get { return this.data; }
      set { this.data = value; }
    }

    [DefaultValue(false)]
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public bool Dirty {
      get { return this.dirty; }
      set {

        // If we're being set to dirty, that implies change 
        // (even if we're already dirty)
        if( this.dirty && this.DocumentChanged != null ) {
          this.DocumentChanged(this, EventArgs.Empty);
        }

        // No need to update caption if we're already at this state
        if( this.dirty == value ) return;
        this.dirty = value;
        SetCaption();
      }
    }

    [DefaultValue(null)]
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public string FileName {
      get { return this.fileName; }
      set {
        this.fileName = value;
        SetCaption();
      }
    }

    [DefaultValue("Untitled")]
    [Category("Behavior")]
    [Description("Gets or sets the default file type, used to determine default file name (\"Untitled\", \"Document\" etc).")]
    public string DefaultFileName {
      get { return this.defaultFileName; }
      set { this.defaultFileName = value; }
    }
    
    [DefaultValue(null)]
    [Category("Behavior")]
    [Description("Gets or sets the default document file extension.")]
    public string FileExtension {
      get { return this.fileExtension; }
      set {
        this.fileExtension = value;
        if( string.IsNullOrEmpty(fileExtension) ) return;

        // Drop leading dot
        while( this.fileExtension.StartsWith(".") ) this.fileExtension = this.fileExtension.Substring(1);
      }
    }

    [DefaultValue(null)]
    [Category("Behavior")]
    [Description("Gets or sets the default document file extension description.")]
    public string FileExtensionDescription {
      get { return this.fileExtensionDesc; }
      set { this.fileExtensionDesc = value; }
    }

    [DefaultValue(null)]
    [Category("Behavior")]
    [Description("e.g \"C# files (*.cs)|*.cs|All files (*.*)|*.*\"")]
    public string Filter {
      get {
        // Cruft up a filter if none has been set at design-time
        if( string.IsNullOrEmpty(this.filter) && !string.IsNullOrEmpty(this.fileExtension) && !DesignMode ) {
          return string.Format("{0} files (*.{1})|*.{1}|All files|*.*", Application.ProductName, fileExtension);
        }
        return this.filter;
      }
      set { this.filter = value; }
    }

    [DefaultValue("")]
    [Category("Appearance")]
    [Description("Gets or sets the default document icon resource Id for the shell from the assembly.")]
    public string IconResourceIndex {
      get { return this.iconResourceIndex; }
      set { this.iconResourceIndex = value; }
    }
    
    // From Ian Griffiths (thanks, Ian!)
    [Browsable(false)]
    public Form HostingForm {
      get {
        if( (host == null) && this.DesignMode ) {
          // See if we're being hosted in VS.NET (or something similar)
          IDesignerHost designer = this.GetService(typeof(IDesignerHost)) as IDesignerHost;
          if( designer != null ) host = designer.RootComponent as Form;
        }

        return host;
      }

      set {
        // Detach from host's Load/Closing event
        if( host != null ) {
          host.Closing -= new CancelEventHandler(host_Closing);
        }

        // Cache the hosting form
        host = value;

        // Attach to host's Load/Closing event
        if( host != null ) {
          // To refresh caption in MDI child case
          host.Load += new EventHandler(host_Load);

          // To handle dirty bit
          host.Closing += new CancelEventHandler(host_Closing);
        }

        // Set the host's caption
        SetCaption();
      }
    }

    [DefaultValue(false)]
    [Category("Behavior")]
    [Description("Gets or sets whether to register the default document file extension with the shell.")]
    public bool RegisterFileExtensionWithShell {
      get { return this.registerFileExtension; }
      set { this.registerFileExtension = value; }
    }

    static void SetMenuItem(ref ToolStripMenuItem thisItem, ToolStripMenuItem newItem, EventHandler click) {
      // Unhook from the old one
      if( thisItem != null ) thisItem.Click -= click;

      // Hook to the new one
      thisItem = newItem;
      if( thisItem != null ) thisItem.Click += click;
    }
    static void SetToolStripButton(ref ToolStripButton thisItem, ToolStripButton newItem, EventHandler click) {
      // Unhook from the old one
      if( thisItem != null ) thisItem.Click -= click;

      // Hook to the new one
      thisItem = newItem;
      if( thisItem != null ) thisItem.Click += click;
    }
    
    [DefaultValue(null)]
    [Category("Menu Strip Integration")]
    [Description("Gets or sets whether to handle File->New menu item.")]
    public ToolStripMenuItem FileNewMenuItem {
      get { return this.fileNewMenuItem; }
      set {
        SetMenuItem(ref this.fileNewMenuItem, value, new EventHandler(fileNew_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Menu Strip Integration")]
    [Description("Gets or sets whether to handle File->Open menu item.")]
    public ToolStripMenuItem FileOpenMenuItem {
      get { return this.fileOpenMenuItem; }
      set {
        SetMenuItem(ref this.fileOpenMenuItem, value, new EventHandler(fileOpen_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Menu Strip Integration")]
    [Description("Gets or sets whether to handle File->Save menu item.")]
    public ToolStripMenuItem FileSaveMenuItem {
      get { return this.fileSaveMenuItem; }
      set {
        SetMenuItem(ref this.fileSaveMenuItem, value, new EventHandler(fileSave_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Menu Strip Integration")]
    [Description("Gets or sets whether to handle File->Save As menu item.")]
    public ToolStripMenuItem FileSaveAsMenuItem {
      get { return this.fileSaveAsMenuItem; }
      set {
        SetMenuItem(ref this.fileSaveAsMenuItem, value, new EventHandler(fileSaveAs_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Menu Strip Integration")]
    [Description("Gets or sets whether to handle File->Save Copy As menu item.")]
    public ToolStripMenuItem FileSaveCopyAsMenuItem {
      get { return this.fileSaveCopyAsMenuItem; }
      set {
        SetMenuItem(ref this.fileSaveCopyAsMenuItem, value, new EventHandler(fileSaveCopyAs_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Menu Strip Integration")]
    [Description("Gets or sets whether to handle File->Close menu item.")]
    public ToolStripMenuItem FileCloseMenuItem {
      get { return this.fileCloseMenuItem; }
      set {
        SetMenuItem(ref this.fileCloseMenuItem, value, new EventHandler(fileClose_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Tool Strip Integration")]
    [Description("Gets or sets whether to handle File->New tool strip button.")]
    public ToolStripButton FileNewToolStripButton {
      get { return this.fileNewToolStripButton; }
      set {
        SetToolStripButton(ref this.fileNewToolStripButton, value, new EventHandler(fileNew_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Tool Strip Integration")]
    [Description("Gets or sets whether to handle File->Open tool strip button.")]
    public ToolStripButton FileOpenToolStripButton {
      get { return this.fileOpenToolStripButton; }
      set {
        SetToolStripButton(ref this.fileOpenToolStripButton, value, new EventHandler(fileOpen_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Tool Strip Integration")]
    [Description("Gets or sets whether to handle File->Save tool strip button.")]
    public ToolStripButton FileSaveToolStripButton {
      get { return this.fileSaveToolStripButton; }
      set {
        SetToolStripButton(ref this.fileSaveToolStripButton, value, new EventHandler(fileSave_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Tool Strip Integration")]
    [Description("Gets or sets whether to handle File->Save As tool strip button.")]
    public ToolStripButton FileSaveAsToolStripButton {
      get { return this.fileSaveAsToolStripButton; }
      set {
        SetToolStripButton(ref this.fileSaveAsToolStripButton, value, new EventHandler(fileSaveAs_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Tool Strip Integration")]
    [Description("Gets or sets whether to handle File->Save Copy As tool strip button.")]
    public ToolStripButton FileSaveCopyAsToolStripButton {
      get { return this.fileSaveCopyAsToolStripButton; }
      set {
        SetToolStripButton(ref this.fileSaveCopyAsToolStripButton, value, new EventHandler(fileSaveCopyAs_Click));
      }
    }

    [DefaultValue(null)]
    [Category("Tool Strip Integration")]
    [Description("Gets or sets whether to handle File->Close tool strip button.")]
    public ToolStripButton FileCloseToolStripButton {
      get { return this.fileCloseToolStripButton; }
      set {
        SetToolStripButton(ref this.fileCloseToolStripButton, value, new EventHandler(fileClose_Click));
      }
    }
    
    static int FileVersion = 0;

    // TODO: Toolbar buttons, too
    //    public ToolBarButton FileOpenButton
    //    {
    //    }

    [DefaultValue(null)]
    [Browsable(false)]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
    public IFormatter Formatter {
      get { return this.formatter; }
      set { this.formatter = value; }
    }

    #endregion

    #region Public Methods
    public bool New() {
      // Check to see if we can close
      if( !Close() ) return false;

      // Reset
      FileVersion++;
      this.Data = null;
      this.Dirty = false;
      this.FileName = null;

      // Let subscribers know there's a new document in town
      if( this.NewDocument != null ) this.NewDocument(this, EventArgs.Empty);

      return true;
    }

    public bool Open() {
      return Open(null);
    }

    public bool Open(string newFileName) {
      // Check if we can close current file
      if( !Close() ) return false;

      // Get a file name
      if( string.IsNullOrEmpty(newFileName) ) {
        // Set up dialog
        this.openFileDialog.DefaultExt = FileExtension;
        this.openFileDialog.Filter = Filter;

        // Get the file to open
        DialogResult res = openFileDialog.ShowDialog(this.host);
        if( res != DialogResult.OK ) return false;

        newFileName = this.openFileDialog.FileName;
      }

      // Read the data
      try {
        Read(newFileName);
      }
      catch( Exception e ) {
        MessageBox.Show(host, "Can't open file: " + newFileName + "\r\n" + e.Message, Application.ProductName);
        return false;
      }

      // Clear dirty bit, cache the file name and set the caption
      this.Dirty = false;
      this.FileName = newFileName;

      // Put the file into the Start->Documents list
      AddToRecentDocs(newFileName);

      // Success
      return true;
    }

    public bool Save() {
      return Save(SaveType.Save);
    }

    public bool SaveAs() {
      return Save(SaveType.SaveAs);
    }

    public bool SaveCopyAs() {
      return Save(SaveType.SaveCopyAs);
    }

    public bool Close() {
      // It's all about the dirty bit...
      if( !this.Dirty ) return true;

      string fileName = ( string.IsNullOrEmpty(this.fileName) ? this.DefaultFileNameAndVersion : this.fileName);

      DialogResult res = MessageBox.Show(host, "Save changes to " + fileName + "?", Application.ProductName, MessageBoxButtons.YesNoCancel);
      switch( res ) {
        case DialogResult.Yes: return Save();
        case DialogResult.No: return true;
        case DialogResult.Cancel: return false;
        default: Debug.Assert(false); return false;
      }
    }

    #endregion

    #region Implementation

    protected bool ShowExtension {
      get {
        // TODO: Check the shell setting
        return true;
      }
    }

    protected void SetCaption() {
      if( this.host == null ) return;
      if( this.DesignMode ) return;

      // Format = "<fileName>*"
      if( this.host.IsMdiChild ) {
        this.host.Text = string.Format("{0}{1}", this.CaptionFileName, (this.Dirty ? "*" : ""));
      }
      // Format = "<appName> - [<fileName>*]"
      else {
        this.host.Text = string.Format("{0} - [{1}{2}]", Application.ProductName, this.CaptionFileName, (this.Dirty ? "*" : ""));
      }
    }

    protected string CaptionFileName {
      get {
        string fileName = this.FileName;
        if( string.IsNullOrEmpty(this.fileName) ) fileName = this.DefaultFileNameAndVersion;
        return (ShowExtension ? Path.GetFileName(fileName) : Path.GetFileNameWithoutExtension(fileName));
      }
    }

    protected string DefaultFileNameAndVersion {
      get { return this.defaultFileName + FileVersion.ToString() + (string.IsNullOrEmpty(this.FileExtension) ? "" : "." + this.FileExtension); }
    }

    protected bool Save(SaveType type) {
      // Get the file name
      string newFileName = this.FileName;
      if( (type == SaveType.SaveAs) || (type == SaveType.SaveCopyAs) || string.IsNullOrEmpty(newFileName) ) {
        // Set up the dialog
        this.saveFileDialog.DefaultExt = FileExtension;
        this.saveFileDialog.Filter = Filter;
        if( !string.IsNullOrEmpty(newFileName) ) {
          this.saveFileDialog.InitialDirectory = Path.GetDirectoryName(newFileName);
          this.saveFileDialog.FileName = Path.GetFileName(newFileName);
        }
        else {
          this.saveFileDialog.FileName = this.DefaultFileNameAndVersion;
        }

        DialogResult res = this.saveFileDialog.ShowDialog(this.host);
        if( res != DialogResult.OK ) return false;
        newFileName = this.saveFileDialog.FileName;
      }

      // Write the data
      try {
        Write(newFileName);
      }
      catch( Exception e ) {
        MessageBox.Show(host, "Can't save file: " + newFileName + "\r\n" + e.Message, Application.ProductName);
        return false;
      }

      if( type != SaveType.SaveCopyAs ) {
        // Clear the dirty bit, cache the new file name and set the caption
        this.Dirty = false;
        this.FileName = newFileName;
      }

      // Put the file into the Start->Documents list
      AddToRecentDocs(newFileName);

      // Success
      return true;
    }

    protected void Write(string fileName) {
      using( Stream stream = new FileStream(fileName, FileMode.Create, FileAccess.Write) ) {
        // Let event subscribers serialize first
        if( this.WriteDocument != null ) {
          SerializeDocumentEventArgs e = new SerializeDocumentEventArgs(fileName, stream);
          this.WriteDocument(this, e);
          if( e.Handled ) return;
        }

        // If subscribers don't serialize, we will
        if( this.formatter != null ) {
          this.formatter.Serialize(stream, Data);
        }
      }
    }

    protected void Read(string fileName) {
      using( Stream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read) ) {
        // Let event subscribers serialize first
        if( this.ReadDocument != null ) {
          SerializeDocumentEventArgs e = new SerializeDocumentEventArgs(fileName, stream);
          this.ReadDocument(this, e);
          if( e.Handled ) return;
        }

        // If subscribers don't serialize, we will
        if( this.formatter != null ) {
          this.Data = formatter.Deserialize(stream);
        }
      }
    }

    void host_Load(object sender, EventArgs e) {
      // Update in case the host is an MDI child
      // NOTE: we do this to avoid setting the MdiParent
      // property on the child 'til the open was a success.
      // Otherwise, child shows up and goes away, which is annoying
      SetCaption();

      if( string.IsNullOrEmpty(this.FileName) ) New();
    }

    private void host_Closing(object sender, CancelEventArgs e) {
      // No need to check if this has already been canceled
      if( e.Cancel ) return;

      // Check to see if we can close the document
      if( !Close() ) e.Cancel = true;
    }

    private void fileNew_Click(object sender, EventArgs e) {
      New();
    }

    private void fileOpen_Click(object sender, EventArgs e) {
      Open();
    }

    private void fileSave_Click(object sender, EventArgs e) {
      Save();
    }

    private void fileSaveAs_Click(object sender, EventArgs e) {
      SaveAs();
    }

    private void fileSaveCopyAs_Click(object sender, EventArgs e) {
      SaveCopyAs();
    }

    private void fileClose_Click(object sender, EventArgs e) {
      // Let Closing event handler decide
      if( this.host != null ) this.host.Close();
    }

    private void CheckRegisterFileExtension() {
      if( !this.registerFileExtension || string.IsNullOrEmpty(this.fileExtension) ) return;

      // Register custom extension with the shell
      string progID = this.fileExtension + "file";
      using( RegistryKey key = Registry.ClassesRoot.CreateSubKey("." + this.fileExtension) ) {
        // Map custom  extension to a ProgID
        key.SetValue(null, progID);
      }

      // Set ProgID description
      if( !string.IsNullOrEmpty(this.FileExtensionDescription) ) {
        using( RegistryKey key = Registry.ClassesRoot.CreateSubKey(progID) ) {
          key.SetValue(null, this.FileExtensionDescription);
        }
      }

      // Register open command with the shell
      string cmdkey = progID + @"\shell\open\command";
      using( RegistryKey key = Registry.ClassesRoot.CreateSubKey(cmdkey) ) {
        // Map ProgID to an Open action for the shell
        key.SetValue(null, string.Format("\"{0}\" \"%L\"", Application.ExecutablePath));
      }

      // TODO: associate icon set via FileDocument component with shell
      // by dumping .ico into a file (if not already there) and registering it
      // with the shell
      // Register document icon with the shell
      if( string.IsNullOrEmpty(this.iconResourceIndex) ) return;
      string icokey = progID + @"\DefaultIcon";
      using( RegistryKey key = Registry.ClassesRoot.CreateSubKey(icokey) ) {
        // Map ProgID to a document icon for the shell
        key.SetValue(null, Application.ExecutablePath + "," + this.iconResourceIndex);
      }
    }

    #endregion

    #region Shell Interop
    enum ShellAddToRecentDocsFlags {
      Pidl = 0x001,
      Path = 0x002, // This is actually PathA
      //PathA = 0x002, // Avoiding these to ensure maximum reach
      //PathW = 0x003,
    }

    // NOTE: The reason for the CharSet.Ansi and limiting to just PathA in
    // the enum is to ensure that the flag and the CharSet match up
    [DllImport("shell32.dll", CharSet = CharSet.Ansi)]
    static extern void SHAddToRecentDocs(ShellAddToRecentDocsFlags flag, string path);

    static void AddToRecentDocs(string path) {
      // Use ANSI for maximum reach
      SHAddToRecentDocs(ShellAddToRecentDocsFlags.Path, path);
    }
    #endregion
  }
}
